# -*- coding: utf-8 -*-
#pylint: disable=invalid-name,missing-function-docstring
#
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Activation functions and helpers in numpy

Here are most of the vector operations and helper functions we use in
numpy.
"""

import numpy as np

#pylint: disable=locally-disabled,no-name-in-module
from scipy.special import expit

from bigml.laminar.constants import LARGE_EXP, MATRIX_PARAMS, \
    VEC_PARAMS, ALPHA, LAMBDA, LEAKY_RELU_CONST


def to_numpy_array(xs):
    if isinstance(xs, np.ndarray):
        return np.copy(xs)
    return np.array(xs, dtype=np.float32)


def softplus(xs):
    x_cpy = to_numpy_array(xs)
    x_cpy[x_cpy < LARGE_EXP] = np.log(np.exp(x_cpy[x_cpy < LARGE_EXP]) + 1)
    return x_cpy


def relu(xs):
    x_cpy = to_numpy_array(xs)
    return x_cpy * (x_cpy > 0)


def softmax(xs):
    x_cpy = to_numpy_array(xs)

    shape0 = 1
    if len(x_cpy.shape) > 1:
        shape0 = x_cpy.shape[0]

    x_cpy = x_cpy.reshape(shape0, -1)

    maxes = np.amax(x_cpy, axis=1)
    maxes = maxes.reshape(maxes.shape[0], 1)

    exps = np.exp(x_cpy - maxes)
    dist = exps / np.sum(exps, axis=1).reshape((-1, 1))

    return dist


def selu(xs):
    x_cpy = to_numpy_array(xs)

    return np.where(x_cpy > 0,
                    LAMBDA * x_cpy,
                    LAMBDA * ALPHA * (np.exp(x_cpy) - 1))

def leaky_relu(xs):
    x_cpy = to_numpy_array(xs)

    return np.maximum(x_cpy, x_cpy * LEAKY_RELU_CONST)


ACTIVATORS = {
    'tanh': np.tanh,
    'sigmoid': expit,
    'softplus': softplus,
    'relu': relu,
    'softmax': softmax,
    'identity': lambda x: x,
    'linear': lambda x: x,
    'swish': lambda x: x * expit(x),
    'mish': lambda x: np.tanh(softplus(x)),
    'relu6': lambda x: np.clip(relu(x), 0, 6),
    'leaky_relu': leaky_relu,
    'selu': selu}


def plus(mat, vec):
    return mat + vec


def dot(mat1, mat2):
    output = []
    for row1 in mat1:
        new_row = []
        for row2 in mat2:
            new_row.append(np.dot(row1, row2).tolist())
        output.append(new_row)
    return output

def batch_norm(X, mean, stdev, shift, scale):
    return scale * (X - mean) / stdev + shift


def init_layer(layer, ftype=np.float64):
    out_layer = {}
    for key in layer:
        if layer[key] is not None:
            if key in MATRIX_PARAMS:
                out_layer[key] = np.array(layer[key], dtype=ftype)
            elif key in VEC_PARAMS:
                out_layer[key] = np.array(layer[key], dtype=ftype)
            else:
                out_layer[key] = layer[key]
        else:
            out_layer[key] = layer[key]

    return out_layer


def init_layers(layers):
    return [init_layer(layer) for layer in layers]


def destandardize(vec, v_mean, v_stdev):
    return vec * v_stdev + v_mean


def to_width(mat, width):
    if width > len(mat[0]):
        ntiles = int(np.ceil(width / float(len(mat[0]))))
    else:
        ntiles = 1

    return np.tile(mat, (1, ntiles))[:, :width]


def add_residuals(residuals, values):
    to_add = to_width(values, len(residuals[0]))
    return to_add + residuals


def sum_and_normalize(youts, is_regression):
    ysums = sum(youts)

    if is_regression:
        return ysums / len(youts)
    return ysums / np.sum(ysums, axis=1).reshape(-1, 1)


def propagate(x_in, layers):
    last_X = identities = to_numpy_array(x_in)

    if any(layer["residuals"] for layer in layers):
        first_identities = not any(layer["residuals"] for layer in layers[:2])
    else:
        first_identities = False

    for i, layer in enumerate(layers):
        w = layer['weights']
        m = layer['mean']
        s = layer['stdev']
        b = layer['offset']
        g = layer['scale']

        afn = layer['activation_function']

        X_dot_w = dot(last_X, w)
        if m is not None and s is not None:
            next_in = batch_norm(X_dot_w, m, s, b, g)
        else:
            next_in = plus(X_dot_w, b)

        if layer['residuals']:
            next_in = add_residuals(next_in, identities)
            last_X = ACTIVATORS[afn](next_in)
            identities = last_X
        else:
            last_X = ACTIVATORS[afn](next_in)

            if first_identities and i == 0:
                identities = last_X

    return last_X



# -*- coding: utf-8 -*-
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import json

from .world import eq_, approx_


def i_create_a_local_forecast(step, input_data):
    """Step: I create a local forecast for <input_data>"""
    input_data = json.loads(input_data)
    step.bigml["local_forecast"] = step.bigml[ \
        "local_time_series"].forecast(input_data)


def the_local_forecast_is(step, local_forecasts):
    """Step: the local forecast is <local_forecasts>"""
    local_forecasts = json.loads(local_forecasts)
    attrs = ["point_forecast", "model"]
    for field_id in local_forecasts:
        forecast = step.bigml["local_forecast"][field_id]
        local_forecast = local_forecasts[field_id]
        eq_(len(forecast), len(local_forecast), msg="forecast: %s" % forecast)
        for index, forecast_item in enumerate(forecast):
            for attr in attrs:
                if isinstance(forecast_item[attr], list):
                    for pos, item in enumerate(forecast_item[attr]):
                        approx_(local_forecast[index][attr][pos],
                                item, precision=5)
                else:
                    eq_(forecast_item[attr], local_forecast[index][attr])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating configuration

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_configuration_steps as config_create

class TestConfiguration:
    """Test for Configuration methods"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating configuration:
            Given I create a configuration from "<configurations>" info
            And I update the configuration name to "<configuration_name>"
            When I wait until the configuration is ready less than <configuration_wait> secs
            Then the configuration name is "<configuration_name>"
            And the configuration contents are "<configurations>"
        """
        show_doc(self.test_scenario1)
        headers = ["configurations", "configuration_wait",
                   "configuration_name"]
        examples = [
            [{
                "dataset": {
                    "name": "Customer FAQ dataset"
                }
            }, '10', {"name": 'my new configuration name'}]]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            config_create.i_create_configuration(
                self, example["configurations"])
            config_create.i_update_configuration(
                self, example["configuration_name"])
            config_create.the_configuration_is_finished_in_less_than(
                self, example["configuration_wait"])
            config_create.i_check_configuration_name(
                self, example["configuration_name"])
            config_create.i_check_configuration_conf(
                self, example["configurations"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY
from bigml.timeseries import TimeSeries

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_


def i_create_a_time_series(step):
    """Step: I create a time series"""
    dataset = world.dataset.get('resource')
    resource = world.api.create_time_series(dataset)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.time_series = resource['object']
    world.time_series_set.append(resource['resource'])


def i_create_a_time_series_with_params(step, data="{}"):
    """Step: I create a time series with params <data>"""
    args = json.loads(data)
    resource = world.api.create_time_series(world.dataset.get('resource'),
                                            args=args)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.time_series = resource['object']
    world.time_series_set.append(resource['resource'])


def the_time_series_is_finished_in_less_than(step, secs):
    """Step: I wait until the time series is ready less than <secs>"""
    world.time_series = wait_until_status_code_is(
        FINISHED, FAULTY, secs, world.time_series)


def create_local_time_series(step):
    """Step: I create a local TimeSeries"""
    step.bigml["local_time_series"] = TimeSeries(world.time_series["resource"],
                                         world.api)


def i_update_time_series_name(step, name):
    """Step: I update the time series name to <name>"""
    resource = world.api.update_time_series(world.time_series['resource'],
                                            {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.time_series = resource['object']


def i_check_time_series_name(step, name):
    """Step: the time series name is <name>"""
    time_series_name = world.time_series['name']
    eq_(name, time_series_name)


def i_export_time_series(step, filename):
    """Step: I export the time series"""
    world.api.export(world.time_series.get('resource'),
                     filename=res_filename(filename))


def i_create_local_time_series_from_file(step, export_file):
    """Step: I create a local time series from file <export_file>"""
    step.bigml["local_time_series"] = TimeSeries(
        res_filename(export_file))


def check_time_series_id_local_id(step):
    """Step: the time series ID and the local time series ID match"""
    eq_(step.bigml["local_time_series"].resource_id,
        world.time_series["resource"])


def clone_time_series(step, time_series):
    """Step: I clone time series"""
    resource = world.api.clone_time_series(time_series,
                                           {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.time_series = resource['object']
    # save reference
    world.time_series_set.append(resource['resource'])


def the_cloned_time_series_is(step, time_series):
    """Checking the time series is a clone"""
    eq_(world.time_series["origin"], time_series)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import os

from bigml.api import HTTP_CREATED
from bigml.api import HTTP_ACCEPTED
from bigml.api import FINISHED
from bigml.api import FAULTY
from bigml.api import get_status
from bigml.topicmodel import TopicModel

from .world import world, res_filename, eq_
from .read_resource_steps import wait_until_status_code_is


def i_create_a_topic_model(step):
    """Step: I create a Topic Model"""
    dataset = world.dataset.get('resource')
    resource = world.api.create_topic_model(
        dataset, {'seed': 'BigML', 'topicmodel_seed': 'BigML'})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.topic_model = resource['object']
    world.topic_models.append(resource['resource'])


def i_create_a_topic_model_from_dataset_list(step):
    """Step: I create a topic model from a dataset list"""
    resource = world.api.create_topic_model(step.bigml["dataset_ids"])
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.topic_model = resource['object']
    world.topic_models.append(resource['resource'])


def i_create_a_topic_model_with_options(step, options):
    """Step: I create a topic model with options <options>"""
    dataset = world.dataset.get('resource')
    options = json.loads(options)
    options.update({'seed': 'BigML',
                    'topicmodel_seed': 'BigML'})
    resource = world.api.create_topic_model(
        dataset, options)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.topic_model = resource['object']
    world.topic_models.append(resource['resource'])


def i_update_topic_model_name(step, name):
    """Step: I update the topic model name to <name>"""
    resource = world.api.update_topic_model(world.topic_model['resource'],
                                            {'name': name})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.topic_model = resource['object']


def wait_until_topic_model_status_code_is(step, code1, code2, secs):
    """Step: I wait until the topic model status code is either
    <code1> or <code2> less than <secs>
    """
    world.topic_model = wait_until_status_code_is(
        code1, code2, secs, world.topic_model)


def the_topic_model_is_finished_in_less_than(step, secs):
    """Steps: I wait until the topic model is ready less than <secs>"""
    wait_until_topic_model_status_code_is(step, FINISHED, FAULTY, secs)


def make_the_topic_model_shared(step):
    """Step: I make the topic model shared """
    resource = world.api.update_topic_model(world.topic_model['resource'],
                                            {'shared': True})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.topic_model = resource['object']


def get_sharing_info(step):
    """Step: I get the topic_model sharing info"""
    world.shared_hash = world.topic_model['shared_hash']
    world.sharing_key = world.topic_model['sharing_key']


def topic_model_from_shared_url(step):
    """Step: I check the topic model status using the topic model\'s
    shared url
    """
    world.topic_model = world.api.get_topic_model("shared/topicmodel/%s" %
                                          world.shared_hash)
    eq_(get_status(world.topic_model)['code'], FINISHED)


def topic_model_from_shared_key(step):
    """Step: I check the topic model status using the topic model\'s
    shared key
    """
    username = os.environ.get("BIGML_USERNAME")
    world.topic_model = world.api.get_topic_model( \
        world.topic_model['resource'],
        shared_username=username, shared_api_key=world.sharing_key)
    eq_(get_status(world.topic_model)['code'], FINISHED)


def i_check_topic_model_name(step, name):
    """Step: the topic model name is <name>"""
    topic_model_name = world.topic_model['name']
    eq_(name, topic_model_name)


def i_create_a_topic_distribution(step, data=None):
    """Step: Create topic distribution """
    if data is None:
        data = "{}"
    topic_model = world.topic_model['resource']
    data = json.loads(data)
    resource = world.api.create_topic_distribution(topic_model, data)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.topic_distribution = resource['object']
    world.topic_distributions.append(resource['resource'])


def i_create_a_local_topic_distribution(step, data=None):
    """Step: I create a local topic distribution"""
    step.bigml["local_topic_distribution"] = \
        step.bigml["local_topic_model"].distribution(json.loads(data))


def i_export_topic_model(step, filename):
    """Step: I export the topic model"""
    world.api.export(world.topic_model.get('resource'),
                     filename=res_filename(filename))


def i_create_local_topic_model_from_file(step, export_file):
    """Step: I create a local topic model from file <export_file>"""
    step.bigml["local_topic_model"] = TopicModel(res_filename(export_file))


def check_topic_model_id_local_id(step):
    """Step: the topic model ID and the local topic model ID match"""
    eq_(step.bigml["local_topic_model"].resource_id,
        world.topic_model["resource"])


def clone_topic_model(step, topic_model):
    """Step: I clone topic model"""
    resource = world.api.clone_topic_model(topic_model,
                                           {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.topic_model = resource['object']
    # save reference
    world.topic_models.append(resource['resource'])


def the_cloned_topic_model_is(step, topic_model):
    """Check cloned topic model"""
    eq_(world.topic_model["origin"], topic_model)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#pylint: disable=locally-disabled,pointless-string-statement
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import json
import os

from zipfile import ZipFile
from bigml.model import Model, cast_prediction
from bigml.logistic import LogisticRegression
from bigml.cluster import Cluster
from bigml.anomaly import Anomaly
from bigml.association import Association
from bigml.multimodel import MultiModel
from bigml.topicmodel import TopicModel
from bigml.deepnet import Deepnet
from bigml.linear import LinearRegression
from bigml.supervised import SupervisedModel
from bigml.local_model import LocalModel
from bigml.fusion import Fusion
from bigml.pca import PCA


from .create_prediction_steps import check_prediction
from .world import world, res_filename, eq_, approx_, ok_


def extract_zip(input_zip):
    """Extracting file names in zip"""
    with ZipFile(input_zip) as zip_handler:
        return {name: zip_handler.read(name) for name in \
            zip_handler.namelist()}


def i_retrieve_a_list_of_remote_models(step, tag):
    """Step: I retrieve a list of remote models tagged with <tag>"""
    world.list_of_models = [ \
        world.api.get_model(model['resource']) for model in
        world.api.list_models(query_string="project=%s;tags__in=%s" % \
            (world.project_id, tag))['objects']]


def i_retrieve_a_list_of_remote_logistic_regressions(step, tag):
    """Step: I retrieve a list of remote logistic regression tagged with
    <tag>
    """
    world.list_of_models = [ \
        world.api.get_logistic_regression(model['resource']) for model in
        world.api.list_logistic_regressions( \
            query_string="project=%s;tags__in=%s" % \
                (world.project_id, tag))['objects']]


def i_retrieve_a_list_of_remote_linear_regressions(step, tag):
    """Step: I retrieve a list of remote linear regression tagged with <tag>"""
    world.list_of_models = [ \
        world.api.get_linear_regression(model['resource']) for model in
        world.api.list_linear_regressions( \
            query_string="project=%s;tags__in=%s" % \
                (world.project_id, tag))['objects']]


def i_create_a_local_model_from_file(step, model_file):
    """Step: I create a local model from a <model_file> file"""
    step.bigml["local_model"] = Model(res_filename(model_file))


def i_create_a_local_deepnet_from_zip_file(step, deepnet_file,
                                           operation_settings=None):
    """Step: I create a local deepnet from a <deepnet_file> file"""
    zipped_files = extract_zip(res_filename(deepnet_file))
    deepnet = json.loads(list(zipped_files.values())[0])
    step.bigml["local_model"] = Deepnet(deepnet,
                                operation_settings=operation_settings)


def i_create_a_local_supervised_model_from_file(step, model_file):
    """Step: I create a local supervised model from a <model_file> file"""
    step.bigml["local_model"] = SupervisedModel(res_filename(model_file))


def i_create_a_local_model(step, pre_model=False):
    """Step: I create a local model"""
    step.bigml["local_model"] = Model(world.model)
    if pre_model:
        step.bigml["local_pipeline"] = step.bigml["local_model"].data_transformations()


def i_create_a_local_fusion(step):
    """Step: I create a local fusion"""
    step.bigml["local_model"] = Fusion(world.fusion['resource'])
    step.bigml["local_ensemble"] = None


def i_create_a_local_supervised_model(step, model_type=None):
    """Step: I create a local supervised model"""
    if model_type is None:
        model_type = "model"
    model = getattr(world, model_type)
    step.bigml["local_model"] = SupervisedModel(model)


def i_create_a_local_bigml_model(step, model_type=None):
    """Step: I create a local BigML model"""
    if model_type is None:
        model_type = "model"
    model = getattr(world, model_type)
    step.bigml["local_model"] = LocalModel(model)


def i_create_a_local_bigml_model_prediction(step, data=None,
    prediction_type=None, **kwargs):
    """Step: I create a local prediction for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    if prediction_type is None:
        prediction_type = "prediction"
    if kwargs is None:
        kwargs = {}
    kwargs.update({"full": True})
    step.bigml["local_%s" % prediction_type] = step.bigml[
        "local_model"].predict(data, **kwargs)


def the_local_bigml_prediction_is(step, value, prediction_type=None, key=None,
    precision=None):
    prediction = step.bigml["local_%s" % prediction_type]
    if key is not None:
        prediction = prediction[key]
    eq_(value, prediction, precision=precision)


def i_create_a_local_prediction_with_confidence(step, data=None,
                                                pre_model=None):
    """Step: I create a local prediction for <data> with confidence"""
    if data is None:
        data = "{}"
    input_data = json.loads(data)
    if pre_model is not None:
        input_data = pre_model.transform([input_data])[0]
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(
        input_data, full=True)


def i_create_a_local_prediction(step, data=None, pre_model=None):
    """Step: I create a local prediction for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    if pre_model is not None:
        data = pre_model.transform([data])[0]
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(data, full=True)


def i_create_a_local_regions_prediction(step, image_file=None):
    """Step: I create a local images prediction for <image_file>"""
    if image_file is None:
        return None
    data = res_filename(image_file)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(data, full=True)
    return step.bigml["local_prediction"]


def i_create_a_local_prediction_op(step, data=None, operating_point=None):
    """Step: I create a local prediction for <data> in operating point
    <operating_point>
    """
    if data is None:
        data = "{}"
    ok_(operating_point is not None)
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict( \
        data, operating_point=operating_point)


def i_create_a_local_ensemble_prediction_op(step, data=None, operating_point=None):
    """Step: I create a local ensemble prediction for <data> in operating
    point <operating_point>
    """
    if data is None:
        data = "{}"
    ok_(operating_point is not None)
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict( \
        data, operating_point=operating_point)


def i_create_local_probabilities(step, data=None):
    """Step: I create local probabilities for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    model = step.bigml["local_model"]
    step.bigml["local_probabilities"] = model.predict_probability(
        data, compact=True)


def i_create_a_local_ensemble_prediction(step, data=None):
    """Step: I create a local ensemble prediction for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict(data)


def i_create_a_local_deepnet_prediction(step, data=None, image_fields=None,
                                        full=False):
    """Step: I create a local deepnet prediction for <data>"""
    if data is None:
        data = "{}"
    if image_fields is None:
        image_fields = []
    data = json.loads(data)
    for field in image_fields:
        if field in data:
            data[field] = res_filename(data[field])
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(data, full=full)


def i_create_a_local_deepnet_prediction_with_op(step, data=None,
                                                operating_point=None):
    """Step: I create a local deepnet prediction with operating point
    for <data>
    """
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict( \
        data, operating_point=operating_point)


def i_create_a_local_median_prediction(step, data=None):
    """Step: I create a local prediction using median for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(data, full=True)


def i_create_a_local_mm_median_batch_prediction(step, data=None):
    """Step: I create a local multimodel batch prediction using median
    for <data>
    """
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].batch_predict(
        [data], to_file=False, use_median=True)[0].predictions[0]['prediction']


def i_create_a_local_proportional_median_prediction(step, data=None):
    """Step: I create a proportional missing strategy local prediction
    using median for <data>
    """
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict( \
        data, missing_strategy=1, full=True)


def i_create_a_local_cluster(step, pre_model=False):
    """Step: I create a local cluster"""
    step.bigml["local_cluster"] = Cluster(world.cluster["resource"])
    if pre_model:
        step.bigml["local_pipeline"] = step.bigml["local_cluster"].data_transformations()


def i_create_a_local_centroid(step, data=None, pre_model=None):
    """Step: I create a local centroid for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    for key, value in list(data.items()):
        if value == "":
            del data[key]
    if pre_model is not None:
        data = pre_model.transform([data])[0]
    step.bigml["local_centroid"] = step.bigml["local_cluster"].centroid(data)


def the_local_centroid_is(step, centroid, distance):
    """Step: the local centroid is <centroid> with distance <distance>"""
    check_prediction(step.bigml["local_centroid"]['centroid_name'], centroid)
    check_prediction(step.bigml["local_centroid"]['distance'], distance)


def i_create_a_local_anomaly(step, pre_model=False):
    """Step: I create a local anomaly detector"""
    step.bigml["local_anomaly"] = Anomaly(world.anomaly["resource"])
    if pre_model:
        step.bigml["local_pipeline"] = step.bigml["local_anomaly"].data_transformations()


def i_create_a_local_anomaly_score(step, input_data, pre_model=None):
    """Step: I create a local anomaly score for <data>"""
    input_data = json.loads(input_data)
    if pre_model is not None:
        input_data = pre_model.transform([input_data])[0]
    step.bigml["local_anomaly_score"] = step.bigml["local_anomaly"].anomaly_score( \
        input_data)


def the_local_anomaly_score_is(step, score):
    """Step: the local anomaly score is <score>"""
    eq_(str(round(step.bigml["local_anomaly_score"], 2)),
        str(round(float(score), 2)))


def i_create_a_local_association(step, pre_model=False):
    """Step: I create a local association"""
    step.bigml["local_association"] = Association(world.association)
    if pre_model:
        step.bigml["local_pipeline"] = step.bigml["local_association"].data_transformations()


def i_create_a_proportional_local_prediction(step, data=None):
    """Step: I create a proportional missing strategy local prediction for
    <data>
    """
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(
        data, missing_strategy=1, full=True)
    step.bigml["local_prediction"] = cast_prediction(step.bigml["local_prediction"],
                                             to="list",
                                             confidence=True)


def i_create_a_prediction_from_a_multi_model(step, data=None):
    """Step: I create a prediction from a multi model for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(data)


def i_create_a_batch_prediction_from_a_multi_model(step, data=None):
    """Step: I create a batch multimodel prediction for <data>"""
    if data is None:
        data = "[{}]"
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].batch_predict(data,
                                                             to_file=False)

def the_batch_mm_predictions_are(step, predictions):
    """Step: the predictions are <predictions>"""
    if predictions is None:
        predictions = "[{}]"
    predictions = json.loads(predictions)
    for index, prediction in enumerate(predictions):
        multivote = step.bigml["local_prediction"][index]
        for mv_prediction in multivote.predictions:
            eq_(mv_prediction['prediction'], prediction)


def the_multiple_local_prediction_is(step, prediction):
    """Step: the multiple local prediction is <prediction>"""
    local_prediction = step.bigml["local_prediction"]
    prediction = json.loads(prediction)
    eq_(local_prediction, prediction)


def the_local_prediction_confidence_is(step, confidence):
    """Step: the local prediction's confidence is <confidence>"""
    if isinstance(step.bigml["local_prediction"], (list, tuple)):
        local_confidence = step.bigml["local_prediction"][1]
    else:
        local_confidence = step.bigml["local_prediction"].get('confidence', \
            step.bigml["local_prediction"].get('probability'))
    local_confidence = round(float(local_confidence), 4)
    confidence = round(float(confidence), 4)
    eq_(local_confidence, confidence)


def the_highest_local_prediction_confidence_is(
    step, input_data, confidence, missing_strategy=None):
    """Step: the highest local prediction's confidence for <input_data> is
    <confidence>"""
    input_data = json.loads(input_data)
    kwargs = {}
    if missing_strategy is not None:
        kwargs.update({"missing_strategy": missing_strategy})
    local_confidence = step.bigml["local_model"].predict_confidence(input_data,
                                                            **kwargs)
    if isinstance(local_confidence, dict):
        local_confidence = round(float(local_confidence["confidence"]), 4)
    else:
        local_confidence = round(float(max([pred["confidence"] for pred in local_confidence])), 4)
    confidence = round(float(confidence), 4)
    eq_(local_confidence, confidence)


def the_local_prediction_is(step, prediction, precision=4):
    """Step: the local prediction is <prediction>"""
    if isinstance(step.bigml["local_prediction"], (list, tuple)):
        local_prediction = step.bigml["local_prediction"][0]
    elif isinstance(step.bigml["local_prediction"], dict):
        local_prediction = step.bigml["local_prediction"]['prediction']
    else:
        local_prediction = step.bigml["local_prediction"]
    if hasattr(world, "local_ensemble") and step.bigml["local_ensemble"] is not None:
        step.bigml["local_model"] = step.bigml["local_ensemble"]
    if (hasattr(step.bigml["local_model"], "regression") and \
            step.bigml["local_model"].regression) or \
            (isinstance(step.bigml["local_model"], MultiModel) and \
            step.bigml["local_model"].models[0].regression):
        local_prediction = round(float(local_prediction), precision)
        prediction = round(float(prediction), precision)
        approx_(local_prediction, float(prediction), precision=precision)
    else:
        if isinstance(local_prediction, str):
            eq_(local_prediction, prediction)
        else:
            if isinstance(prediction, str):
                prediction = float(prediction)
            eq_(round(local_prediction, precision),
                round(float(prediction), precision))


def the_local_regions_prediction_is(step, prediction):
    """Step: the local regions prediction is <prediction>"""
    prediction = json.loads(prediction)
    eq_(prediction, step.bigml["local_prediction"])


def the_local_probabilities_are(step, prediction):
    """Step: the local probabilities are <prediction>"""
    local_probabilities = step.bigml["local_probabilities"]
    expected_probabilities = [float(p) for p in json.loads(prediction)]

    for local, expected in zip(local_probabilities, expected_probabilities):
        approx_(local, expected, precision=4)


def the_local_ensemble_prediction_is(step, prediction):
    """Step: the local ensemble prediction is <prediction>"""
    if isinstance(step.bigml["local_prediction"], (list, tuple)):
        local_prediction = step.bigml["local_prediction"][0]
    elif isinstance(step.bigml["local_prediction"], dict):
        local_prediction = step.bigml["local_prediction"]['prediction']
    else:
        local_prediction = step.bigml["local_prediction"]
    if step.bigml["local_ensemble"].regression:
        approx_(local_prediction, float(prediction), precision=5)
    else:
        eq_(local_prediction, prediction)


def the_local_probability_is(step, probability):
    """Step: the local probability is <probability>"""
    local_probability = step.bigml["local_prediction"]["probability"]
    if isinstance(probability, str):
        probability = float(probability)
    eq_(local_probability, probability, precision=4)


def eq_local_and_remote_probability(step):
    """Step: check local and remote probability"""
    local_probability = round(step.bigml["local_prediction"]["probability"], 3)
    remote_probability = round(world.prediction["probability"], 3)
    approx_(local_probability, remote_probability)


def i_create_a_local_multi_model(step):
    """Step: I create a local multi model"""
    step.bigml["local_model"] = MultiModel(world.list_of_models)
    step.bigml["local_ensemble"] = None


def i_create_a_batch_prediction(step, input_data_list, directory):
    """Step: I create a batch prediction for <input_data_list> and save it
    in <directory>
    """
    if len(directory) > 0 and not os.path.exists(directory):
        os.makedirs(directory)
    input_data_list = json.loads(input_data_list)
    ok_(isinstance(input_data_list, list))
    step.bigml["local_model"].batch_predict(input_data_list, directory)


def i_combine_the_votes(step, directory):
    """Step: I combine the votes in <directory>"""
    world.votes = step.bigml["local_model"].batch_votes(directory)


def the_plurality_combined_prediction(step, predictions):
    """Step: the plurality combined predictions are <predictions>"""
    predictions = json.loads(predictions)
    for i, votes_row in enumerate(world.votes):
        combined_prediction = votes_row.combine()
        check_prediction(combined_prediction, predictions[i])


def the_confidence_weighted_prediction(step, predictions):
    """Step: the confidence weighted predictions are <predictions>"""
    predictions = json.loads(predictions)
    for i, votes_row in enumerate(world.votes):
        combined_prediction = votes_row.combine(1)
        eq_(combined_prediction, predictions[i])


def i_create_a_local_logistic_model(step, pre_model=False):
    """Step: I create a local logistic regression model"""
    step.bigml["local_model"] = LogisticRegression(world.logistic_regression)
    if pre_model:
        step.bigml["local_pipeline"] = step.bigml[
            "local_model"].data_transformations()
    if hasattr(world, "local_ensemble"):
        step.bigml["local_ensemble"] = None


def i_create_a_local_deepnet(step):
    """Step: I create a local deepnet model"""
    step.bigml["local_model"] = Deepnet({"resource": world.deepnet['resource'],
                                 "object": world.deepnet})
    if hasattr(world, "local_ensemble"):
        step.bigml["local_ensemble"] = None


def i_create_a_local_topic_model(step):
    """Step: I create a local topic model"""
    step.bigml["local_topic_model"] = TopicModel(world.topic_model)


def the_topic_distribution_is(step, distribution):
    """Step: the topic distribution is <distribution>"""
    eq_(json.loads(distribution),
        world.topic_distribution['topic_distribution']['result'])


def the_local_topic_distribution_is(step, distribution):
    """Step: the local topic distribution is <distribution>"""
    distribution = json.loads(distribution)
    for index, topic_dist in enumerate(step.bigml["local_topic_distribution"]):
        approx_(topic_dist["probability"], distribution[index])


def the_association_set_is_like_file(step, filename):
    """Step: the association set is like file <filename>"""
    filename = res_filename(filename)
    result = world.association_set.get("association_set",{}).get("result", [])
    """ Uncomment if different text settings are used
    with open(filename, "w") as filehandler:
        json.dump(result, filehandler)
    """
    with open(filename) as filehandler:
        file_result = json.load(filehandler)
    eq_(result, file_result)


def i_create_a_local_association_set(step, data, pre_model=None):
    """Step: I create a local association set"""
    data = json.loads(data)
    if pre_model is not None:
        data = pre_model.transform([data])[0]
    step.bigml["local_association_set"] = step.bigml["local_association"].association_set( \
        data)


def the_local_association_set_is_like_file(step, filename):
    """Step: the local association set is like file <filename>"""
    filename = res_filename(filename)
    """ Uncomment if different text settings are used
    with open(filename, "w") as filehandler:
        json.dump(result, filehandler)
    """
    with open(filename) as filehandler:
        file_result = json.load(filehandler)
        for index, result in enumerate(file_result):
            approx_(result['score'], step.bigml["local_association_set"][
                index]['score'])
            eq_(result['rules'],
                step.bigml["local_association_set"][index]['rules'])


def i_create_a_local_prediction_op_kind(step, data=None, operating_kind=None):
    """Step: I create a local prediction for <data> in operating kind
    <operating_kind>
    """
    if data is None:
        data = "{}"
    ok_(operating_kind is not None)
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict( \
        data, operating_kind=operating_kind)


def i_create_a_local_ensemble_prediction_op_kind( \
        step, data=None, operating_kind=None):
    """Step: I create a local ensemble prediction for <data> in operating
    kind <operating_kind>"""
    if data is None:
        data = "{}"
    ok_(operating_kind is not None)
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_ensemble"].predict( \
        data, operating_kind=operating_kind)


def i_create_a_local_deepnet_prediction_op_kind( \
        step, data=None, operating_kind=None):
    """Step: I create a local deepnet for <data> in operating kind
    <operating_kind>
    """
    if data is None:
        data = "{}"
    ok_(operating_kind is not None)
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict( \
        data, operating_kind=operating_kind)


def i_create_a_local_logistic_prediction_op_kind( \
        step, data=None, operating_kind=None):
    """Step: I create a local logistic regression for <data> in operating
    kind <operating_kind>
    """
    if data is None:
        data = "{}"
    ok_(operating_kind is not None)
    data = json.loads(data)
    step.bigml["local_prediction"] = step.bigml["local_model"].predict( \
        data, operating_kind=operating_kind)


def create_local_pca(step, pre_model=False):
    """Step: I create a local PCA"""
    step.bigml["local_pca"] = PCA(world.pca["resource"])
    if pre_model:
        step.bigml["local_pipeline"] = step.bigml["local_pca"].data_transformations()


def i_create_a_local_linear(step):
    """Step: I create a local linear regression"""
    step.bigml["local_model"] = LinearRegression(world.linear_regression["resource"])


def i_create_a_local_projection(step, data=None, pre_model=None):
    """Step: I create a local projection for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    if pre_model is not None:
        data = pre_model.transform([data])[0]
    for key, value in list(data.items()):
        if value == "":
            del data[key]
    step.bigml["local_projection"] = step.bigml["local_pca"].projection(data, full=True)
    for name, value in list(step.bigml["local_projection"].items()):
        step.bigml["local_projection"][name] = round(value, 5)


def i_create_a_local_linear_prediction(step, data=None):
    """Step: I create a local linear regression prediction for <data>"""
    if data is None:
        data = "{}"
    data = json.loads(data)
    for key, value in list(data.items()):
        if value == "":
            del data[key]
    step.bigml["local_prediction"] = step.bigml["local_model"].predict(data, full=True)
    for name, value in list(step.bigml["local_prediction"].items()):
        if isinstance(value, float):
            step.bigml["local_prediction"][name] = round(value, 5)


def the_local_projection_is(step, projection):
    """Step: checking the local projection"""
    if projection is None:
        projection = "{}"
    projection = json.loads(projection)
    eq_(len(list(projection.keys())), len(list(step.bigml["local_projection"].keys())))
    for name, _ in list(projection.items()):
        eq_(step.bigml["local_projection"][name], projection[name],
            msg="local: %s, %s - expected: %s" % ( \
                name, step.bigml["local_projection"][name], projection[name]))



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2012-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
import os

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY
from bigml.api import get_status
from bigml.cluster import Cluster

from .read_resource_steps import wait_until_status_code_is
from .world import world, res_filename, eq_


def i_create_a_cluster(step, shared=None):
    """Step: I create a cluster"""
    if shared is None or world.shared.get("cluster", {}).get(shared) is None:
        dataset = world.dataset.get('resource')
        resource = world.api.create_cluster(
            dataset, {'seed': 'BigML',
                      'cluster_seed': 'BigML',
                      'k': 8})
        world.status = resource['code']
        eq_(world.status, HTTP_CREATED)
        world.location = resource['location']
        world.cluster = resource['object']
        world.clusters.append(resource['resource'])


def i_create_a_cluster_from_dataset_list(step):
    """Step: I create a cluster from a dataset list"""
    resource = world.api.create_cluster(step.bigml["dataset_ids"])
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.cluster = resource['object']
    world.clusters.append(resource['resource'])


def i_create_a_cluster_with_options(step, options):
    """Step: I create a cluster with options <options>"""
    dataset = world.dataset.get('resource')
    options = json.loads(options)
    options.update({'seed': 'BigML',
                    'cluster_seed': 'BigML',
                    'k': 8})
    resource = world.api.create_cluster(
        dataset, options)
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.cluster = resource['object']
    world.clusters.append(resource['resource'])


def wait_until_cluster_status_code_is(step, code1, code2, secs):
    """Step: I wait until the cluster status code is either <code1> or
    <code2> less than <secs>"""
    world.cluster = wait_until_status_code_is(
        code1, code2, secs, world.cluster)


def the_cluster_is_finished_in_less_than(step, secs, shared=None):
    """Step: I wait until the cluster is ready less than <secs>"""
    if shared is None or world.shared.get("cluster", {}).get(shared) is None:
        wait_until_cluster_status_code_is(step, FINISHED, FAULTY, secs)
        if shared is not None:
            if "cluster" not in world.shared:
                world.shared["cluster"] = {}
            world.shared["cluster"][shared] = world.cluster
    else:
        world.cluster = world.shared["cluster"][shared]
        print("Reusing %s" % world.cluster["resource"])


def make_the_cluster_shared(step):
    """Step: I make the cluster shared"""
    resource = world.api.update_cluster(world.cluster['resource'],
                                      {'shared': True})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.cluster = resource['object']


def get_sharing_info(step):
    """Step: I get the cluster sharing info"""
    world.shared_hash = world.cluster['shared_hash']
    world.sharing_key = world.cluster['sharing_key']


def cluster_from_shared_url(step):
    """Step: I check the cluster status using the model's shared url"""
    world.cluster = world.api.get_cluster("shared/cluster/%s" % world.shared_hash)
    eq_(get_status(world.cluster)['code'], FINISHED)


def cluster_from_shared_key(step):
    """Step: I check the cluster status using the model's shared key"""
    username = os.environ.get("BIGML_USERNAME")
    world.cluster = world.api.get_cluster(world.cluster['resource'],
        shared_username=username, shared_api_key=world.sharing_key)
    eq_(get_status(world.cluster)['code'], FINISHED)


def closest_in_cluster(step, reference, closest):
    """Step: the data point in the cluster closest to <reference> is <closest>"""
    local_cluster = step.bigml["local_cluster"]
    reference = json.loads(reference)
    closest = json.loads(closest)
    result = local_cluster.closest_in_cluster( \
        reference, number_of_points=1)["closest"][0]
    result = json.loads(json.dumps(result))
    eq_(closest, result)


def closest_centroid_in_cluster(step, reference, closest_id):
    """Step: the centroid in the cluster closest to <reference> is
    <closest_id>
    """
    local_cluster = step.bigml["local_cluster"]
    reference = json.loads(reference)
    result = local_cluster.sorted_centroids( \
        reference)
    result = result["centroids"][0]["centroid_id"]
    eq_(closest_id, result)

def i_export_cluster(step, filename):
    """Step: I export the cluster"""
    world.api.export(world.cluster.get('resource'),
                     filename=res_filename(filename))


def i_create_local_cluster_from_file(step, export_file):
    """Step: I create a local cluster from file <export_file>"""
    step.bigml["local_cluster"] = Cluster(res_filename(export_file))


def check_cluster_id_local_id(step):
    """Step: the cluster ID and the local cluster ID match"""
    eq_(step.bigml["local_cluster"].resource_id, world.cluster["resource"])


def clone_cluster(step, cluster):
    """Step: I clone cluster"""
    resource = world.api.clone_cluster(cluster,
                                       {'project': world.project_id})
    # update status
    world.status = resource['code']
    world.location = resource['location']
    world.cluster = resource['object']
    # save reference
    world.clusters.append(resource['resource'])


def the_cloned_cluster_is(step, cluster):
    """Checking the cluster is a clone"""
    eq_(world.cluster["origin"], cluster)



# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utilities for webhooks

"""
import json
import hmac

try:
    from hashlib import sha1
except ImportError:
    import sha
    sha1 = sha.sha


SORTING_SEQUENCE = ["timestamp", "message", "resource", "event"]


def dict_to_msg(obj):
    """Builds a representation of the dict object in a specific key sequence"""
    pair_list = []
    for key in SORTING_SEQUENCE:
        pair_list.append("'%s': '%s'" % (key, obj.get(key)))
    return "{%s}" % ", ".join(pair_list)


def compute_signature(msg, secret, encoding="utf-8"):
    """Computes the signature used by BigML when issuing the webhook call"""
    return hmac.new(
        secret.encode(encoding),
        msg=msg.encode(encoding),
        digestmod=sha1
    ).hexdigest()


def check_signature(request, secret):
    """Checks the signature when the webhook has been given one"""
    sig_header = request.meta['HTTP_X_BIGML_SIGNATURE'].replace('sha1=', '')
    payload = request.body
    computed_sig = compute_signature(payload, secret)
    if sig_header == computed_sig:
        return True
    # code for old version of the msg hash
    payload = dict_to_msg(json.loads(payload))
    computed_sig = compute_signature(payload, secret)
    if sig_header == computed_sig:
        return True
    return False



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Testing local dataset transformations

"""
import os
import json

from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import compare_pipeline_steps as pipeline_compare
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_anomaly_steps as anomaly_create
from . import create_model_steps as model_create
from . import create_ensemble_steps as ensemble_create
from . import create_linear_steps as linear_create
from . import create_prediction_steps as prediction_create
from . import compare_predictions_steps as prediction_compare


class TestLocalPipeline:
    """Testing local Pipeline methods"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario 1: Successfully creating a local pipeline from a model and anomaly detector:
            Given I expand the zip file "<pipeline_file>" that contain "<models_list>"
            And I create a local pipeline for "<models_list>" named "<name>"
            Then the transformed data for "<input_data>" is "<output_data>"
        """
        show_doc(self.test_scenario1)
        headers = ["pipeline_file", "models_list", "name", "input_data",
                   "output_data"]
        examples = [
            ['bigml/tests/pipeline3.zip',
             '["anomaly/631a6a968f679a2d2d000319",'
             ' "model/631a6a6f8f679a2d31000445"]',
             "pipeline3",
             '{"plasma glucose": 120, "age": 30, "bmi": 46}',
             '{"plasma glucose": 120, "age": 30, "glucose half": 60,'
             ' "age_range": "2nd third", "bmi": 46,'
             ' "score": 0.85456,'
             ' "prediction": "false", "probability": 0.6586746586746587}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            pipeline_compare.i_expand_file_with_models_list(
                self, example["pipeline_file"], example["models_list"])
            pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, example["models_list"], example["name"],
                storage=os.path.splitext(example["pipeline_file"])[0])
            pipeline_compare.the_pipeline_transformed_data_is(
                self, example["input_data"], example["output_data"])

    def test_scenario2(self):
        """
        Scenario 2: Successfully creating a local pipeline from two BMLPipelines
            Given I expand the zip file "<pipeline_file>" that contain "<models_list>"
            And I create a local pipeline for "<model1>" named "<name1>"
            And I create a local pipeline for "<model2>" named "<name2>"
            And I create a local pipeline "<name3>" for both pipelines
            Then the transformed data for "<input_data>" is "<output_data>"
        """
        show_doc(self.test_scenario2)
        headers = ["pipeline_file", "models_list", "model1", "name1",
                   "model2", "name2", "name", "input_data", "output_data"]
        examples = [
            ['bigml/tests/pipeline3.zip',
             '["anomaly/631a6a968f679a2d2d000319",'
             ' "model/631a6a6f8f679a2d31000445"]',
             '["model/631a6a6f8f679a2d31000445"]',
             "pipeline1",
             '["anomaly/631a6a968f679a2d2d000319"]',
             "pipeline2",
             "pipeline3",
             '{"plasma glucose": 120, "age": 30, "bmi": 46}',
             '{"plasma glucose": 120, "age": 30, "glucose half": 60,'
             ' "age_range": "2nd third", "bmi": 46,'
             ' "score": 0.85456,'
             ' "prediction": "false", "probability": 0.6586746586746587}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            pipeline_compare.i_expand_file_with_models_list(
                self, example["pipeline_file"], example["models_list"])
            pipe1 = pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, example["model1"], example["name1"],
                storage=os.path.splitext(example["pipeline_file"])[0])
            pipe2 = pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, example["model2"], example["name2"],
                storage=os.path.splitext(example["pipeline_file"])[0])
            pipeline_compare.i_create_composed_pipeline(self, [pipe1, pipe2],
                                                        example["name"])
            pipeline_compare.the_pipeline_transformed_data_is(
                self, example["input_data"], example["output_data"])

    def test_scenario3(self):
        """
        Scenario: Successfully comparing remote and local predictions
                  with raw date input for linear regression:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a linear regression
            And I wait until the linear regression is ready
            less than <model_wait> secs
            And I create a local pipeline for the linear regression named "<name>"
            When I create a prediction for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And the prediction in the transformed data for "<input_data>" is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "name"]
        examples = [
            ['data/dates2.csv', '20', '20', '25',
             '{"time-1": "1910-05-08T19:10:23.106", "cat-0":"cat2"}',
             '000002',  -0.01284, "pipeline1"],
            ['data/dates2.csv', '20', '20', '25',
             '{"time-1": "1920-06-30T20:21:20.320", "cat-0":"cat1"}',
             '000002', -0.09459, "pipeline2"],
            ['data/dates2.csv', '20', '20', '25',
             '{"time-1": "1932-01-30T19:24:11.440",  "cat-0":"cat2"}',
             '000002', -0.02259, "pipeline3"],
            ['data/dates2.csv', '20', '20', '25',
             '{"time-1": "1950-11-06T05:34:05.252", "cat-0":"cat1"}',
             '000002', -0.06754, "pipeline4"],
            ['data/dates2.csv', '20', '20', '25',
             '{"time-1": "2001-01-05T23:04:04.693", "cat-0":"cat2"}',
             '000002', 0.05204, "pipeline5"],
            ['data/dates2.csv', '20', '20', '25',
             '{"time-1": "2011-04-01T00:16:45.747", "cat-0":"cat2"}',
             '000002', 0.05878, "pipeline6"]]
        show_doc(self.test_scenario3)

        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            linear_create.i_create_a_linear_regression(
                self, shared=example["data"])
            linear_create.the_linear_regression_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, [world.linear_regression["resource"]], example["name"])
            prediction_create.i_create_a_linear_prediction(
                self, example["input_data"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            pipeline_compare.the_pipeline_result_key_is(
                self, example["input_data"], "prediction",
                example["prediction"])

    def test_scenario4(self):
        """
        Scenario: Successfully comparing remote and local predictions
                  with raw date input for deepnet:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a deepnet
            And I wait until the deepnet is ready less than <model_wait> secs
            And I create a local pipeline for the deepnet named "<name>"
            When I create a prediction for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And the prediction in the transformed data for "<input_data>" is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "name"]
        examples = [
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "1910-05-08T19:10:23.106", "cat-0":"cat2"}',
             '000002', -0.02616, "pipeline1"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "2011-04-01T00:16:45.747", "cat-0":"cat2"}',
             '000002', 0.13352, "pipeline2"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "1969-W29-1T17:36:39Z", "cat-0":"cat1"}',
             '000002', 0.10071, "pipeline3"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "1920-06-45T20:21:20.320", "cat-0":"cat1"}',
             '000002', 0.10071, "pipeline4"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "2001-01-05T23:04:04.693", "cat-0":"cat2"}',
             '000002', 0.15235, "pipeline5"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "1950-11-06T05:34:05.602", "cat-0":"cat1"}',
             '000002', -0.07686, "pipeline6"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "1932-01-30T19:24:11.440",  "cat-0":"cat2"}',
             '000002', 0.0017, "pipeline7"],
            ['data/dates2.csv', '20', '45', '160',
             '{"time-1": "Mon Jul 14 17:36 +0000 1969", "cat-0":"cat1"}',
             '000002', 0.10071, "pipeline8"]]
        show_doc(self.test_scenario4)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            deepnet_shared = "%s_no_sug" % example["data"]
            model_create.i_create_a_no_suggest_deepnet(
                self, shared=deepnet_shared)
            model_create.the_deepnet_is_finished_in_less_than(
                self, example["model_wait"], shared=deepnet_shared)
            pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, [world.deepnet["resource"]], example["name"])
            prediction_create.i_create_a_deepnet_prediction(
                self, example["input_data"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"],
                precision=4)
            pipeline_compare.the_pipeline_result_key_is(
                self, example["input_data"], "prediction",
                example["prediction"], precision=4)

    def test_scenario5(self):
        """
        Scenario: Successfully comparing remote and local predictions
                  with raw input for deepnets with images:
            Given I create an annotated images data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a deepnet with parms <model_conf>
            And I wait until the deepnet is ready
            less than <time_3> secs
            And I create a local pipeline for the deepnet named "<name>"
            When I create a prediction for "<input_data>"
            Then the prediction for "<objective>" is "<prediction>"
            When I create a prediction for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And the prediction in the transformed data for "<input_data>" is "<prediction>"
            And the probability in the transformed data for "<input_data>" is "<probability>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "model_conf", "image_fields",
                   "name"]
        examples = [
            ['data/images/metadata.json', '500', '500', '600',
             '{"image_id": "data/fruits1e.jpg", "label":"f1"}',
             '100003', {"objective_field": "100003",
                        "number_of_hidden_layers": 1,
                        "suggest_structure": False,
                        "missing_numerics": True,
                        "max_training_time": 100,
                        "hidden_layers": [{
                            "activation_function": "tanh",
                            "number_of_nodes": 10}]},
             ['image_id'], "pipeline1"]]
        show_doc(self.test_scenario5)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_create_annotated_source(
                self,
                example["data"],
                args={"image_analysis": {"enabled": False,
                                         "extracted_features": []}})
            source_create.the_source_is_finished(
                self, example["source_wait"])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            model_create.i_create_a_deepnet_with_objective_and_params(
                self, example["objective_id"],
                json.dumps(example["model_conf"]))
            model_create.the_deepnet_is_finished_in_less_than(
                self, example["model_wait"])
            pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, [world.deepnet["resource"]], example["name"])
            prediction_create.i_create_a_deepnet_prediction(
                self, example["input_data"], example["image_fields"])
            prediction = world.prediction["output"]
            probability = world.prediction["probability"]
            pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, [world.deepnet["resource"]], example["name"])
            pipeline_compare.the_pipeline_result_key_is(
                self, example["input_data"], "prediction", prediction,
                precision=4)
            pipeline_compare.the_pipeline_result_key_is(
                self, example["input_data"], "probability", probability,
                precision=2)

    def test_scenario6(self):
        """
        Scenario: Successfully comparing remote and local anomaly scores
                  with raw input for dataset with images:
            Given I create an annotated images data source uploading a "<data>" file and <extracted_features>

            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create an anomaly detector
            And I wait until the anomaly is ready
            less than <anomaly_wait> secs
            And I create a local pipeline for the anomaly detector named "<name>"
            When I create an anomaly score for "<input_data>"
            Then the anomaly score is "<score>"
            And the anomaly score in the transformed data for "<input_data>" is "<score>"
        """
        headers = ["data", "extracted_features", "source_wait", "dataset_wait",
                   "anomaly_wait", "input_data", "score", "name"]
        examples = [
            ['data/images/fruits_hist.zip',
             ["dimensions", "average_pixels", "level_histogram",
              "histogram_of_gradients", ["wavelet_subbands", 5],
              ["pretrained_cnn", "mobilenetv2"]],
              '500', '500', '600',
             '{"image_id": "data/fruits1e.jpg"}', 0.39902, "pipeline1"]]
        show_doc(self.test_scenario6)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file_with_args(
                self,
                example["data"],
                args=json.dumps({"image_analysis": {
                    "enabled": True,
                    "extracted_features": example["extracted_features"]}}))
            source_create.the_source_is_finished(
                self, example["source_wait"])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            anomaly_create.i_create_an_anomaly(self)
            anomaly_create.the_anomaly_is_finished_in_less_than(
                self, example["anomaly_wait"])
            pipeline_compare.i_create_a_local_pipeline_from_models_list(
                self, [world.anomaly["resource"]], example["name"])
            prediction_create.i_create_an_anomaly_score(
                self, example["input_data"])
            score = world.anomaly_score["score"]
            prediction_create.the_anomaly_score_is(
                self, world.anomaly_score["score"])
            pipeline_compare.the_pipeline_result_key_is(
                self, example["input_data"], "score", score)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2019-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating Linear Regression

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_linear_steps as linear_create
from . import create_prediction_steps as prediction_create
from . import create_batch_prediction_steps as batch_pred_create

class TestLinearRegression:
    """Testing Linear Regression methods"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating a linear regression from a dataset:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a linear regression from a dataset
            And I wait until the linear regression is ready less than <model_wait> secs
            And I update the linear regression name to "<linear_name>"
            When I wait until the linear regression is ready less than <model_wait> secs
            Then the linear regression name is "<linear_name>"
        """
        show_doc(self.test_scenario1)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "linear_name"]
        examples = [
            ['data/grades.csv', '100', '100', '200', 'my new linear regression name']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            linear_create.i_create_a_linear_regression_from_dataset(self)
            linear_create.the_linear_regression_is_finished_in_less_than(
                self, example["model_wait"])
            linear_create.i_update_linear_regression_name(
                self, example["linear_name"])
            linear_create.the_linear_regression_is_finished_in_less_than(
                self, example["model_wait"])
            linear_create.i_check_linear_name(self, example["linear_name"])

        print("\nEnd of tests in: %s\n-------------------\n" % __name__)

    def test_scenario2(self):
        """
        Scenario: Successfully creating a prediction from linear regression:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a pca
            And I wait until the linear regression is ready less than <model_wait> secs
            When I create a prediction for "<input_data>"
            Then the prediction is "<prediction>"
        """
        show_doc(self.test_scenario2)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "model_conf"]
        examples = [
            ['data/grades.csv', '30', '30', '50',
             '{"000000": 0.5, "000001": 1, "000002": 1, "000003": 1}',
             "000005", '2.27312', '{}'],
            ['data/grades.csv', '30', '30', '50',
             '{"000000": 0.5, "000001": 1, "000002": 1, "000003": 1}',
             "000005", '8.19619', '{"bias": false}'],
            ['data/dates.csv', '30', '30', '30',
             '{"test-num1": 23, "test-num2" : 54, "test-date.day-of-month":2, '
             '"test-date.month":12, "test-date.day-of-week": 2, '
             '"test-date.year": 2012}', "000003", '48.27679',
             '{"bias": false}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"])
            dataset_create.i_create_a_dataset(self)
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"])
            linear_create.i_create_a_linear_regression_with_objective_and_params(
                self, example["objective_id"], example["model_conf"])
            linear_create.the_linear_regression_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_create.i_create_a_linear_prediction(
                self, example["input_data"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])

    def test_scenario3(self):
        """
        Scenario: Successfully creating a batch prediction from a linear regression:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a linear regression
            And I wait until the linear regression is ready less than <model_wait> secs
            When I create a batch prediction for the dataset with the linear regression
            And I wait until the batch predictin is ready less than <batch_wait> secs
            And I download the created predictions file to "<local_file>"
            Then the batch prediction file is like "<predictions_file>"
        """
        show_doc(self.test_scenario3)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "batch_wait", "local_file", "predictions_file"]
        examples = [
            ['data/grades.csv', '30', '30', '50', '50',
             'tmp/batch_predictions.csv', 'data/batch_predictions_linear.csv']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            linear_create.i_create_a_linear_regression_from_dataset(self)
            linear_create.the_linear_regression_is_finished_in_less_than(
                self, example["model_wait"])
            batch_pred_create.i_create_a_linear_batch_prediction(self)
            batch_pred_create.the_batch_prediction_is_finished_in_less_than(
                self, example["batch_wait"])
            batch_pred_create.i_download_predictions_file(
                self, example["local_file"])
            batch_pred_create.i_check_predictions(
                self, example["predictions_file"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Comparing remote and local predictions

"""
import json

from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_anomaly_steps as anomaly_create
from . import create_model_steps as model_create
from . import create_ensemble_steps as ensemble_create
from . import create_linear_steps as linear_create
from . import create_prediction_steps as prediction_create
from . import compare_predictions_steps as prediction_compare


class TestComparePrediction:
    """Test local and remote predictions"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully comparing predictions for deepnets:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a deepnet with objective "<objective_id>" and "<model_conf>"
            And I wait until the deepnet is ready less than <model_wait> secs
            And I create a local deepnet
            When I create a prediction for "<input_data>"
            Then the prediction for "<objective>" is "<prediction>"
            And I create a local prediction for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "model_conf"]
        examples = [
            ['data/iris.csv', '30', '50', '60', '{"petal width": 4}', '000004',
             'Iris-virginica', '{}'],
            ['data/iris.csv', '30', '50', '60',
             '{"sepal length": 4.1, "sepal width": 2.4}', '000004',
             'Iris-versicolor', '{}'],
            ['data/iris_missing2.csv', '30', '50', '60', '{}', '000004',
             'Iris-versicolor', '{}'],
            ['data/grades.csv', '30', '50', '60', '{}', '000005', 55.6560,
             '{}'],
            ['data/spam.csv', '30', '50', '60', '{}', '000000', 'ham', '{}']]
        show_doc(self.test_scenario1)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_deepnet_with_objective_and_params(
                self, example["objective_id"], example["model_conf"])
            model_create.the_deepnet_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_compare.i_create_a_local_deepnet(self)
            prediction_create.i_create_a_deepnet_prediction(
                self, example["input_data"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"],
                precision=3)
            prediction_compare.i_create_a_local_deepnet_prediction(
                self, example["input_data"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"], precision=3)

    def test_scenario2(self):
        """
        Scenario: Successfully comparing predictions in operating points for models:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a model
            And I wait until the model is ready less than <model_wait> secs
            And I create a local model
            When I create a prediction for "<input_data>" in "<operating_point>"
            Then the prediction for "<objective>" is "<prediction>"
            And I create a local prediction for "<input_data>" in "<operating_point>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "prediction", "operating_point",
                   "objective_id"]
        examples = [
            ['data/iris.csv', '10', '50', '50', '{"petal width": 4}',
             'Iris-setosa',
             {"kind": "probability", "threshold": 0.1,
              "positive_class": "Iris-setosa"}, "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal width": 4}',
             'Iris-versicolor',
             {"kind": "probability", "threshold": 0.9,
              "positive_class": "Iris-setosa"}, "000004"],
            ['data/iris.csv', '10', '50', '50',
             '{"sepal length": 4.1, "sepal width": 2.4}',  'Iris-setosa',
             {"kind": "confidence", "threshold": 0.1,
              "positive_class": "Iris-setosa"}, "000004"],
            ['data/iris.csv', '10', '50', '50',
             '{"sepal length": 4.1, "sepal width": 2.4}', 'Iris-versicolor',
             {"kind": "confidence", "threshold": 0.9,
              "positive_class": "Iris-setosa"}, "000004"]]
        show_doc(self.test_scenario2)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_model(self, shared=example["data"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            prediction_compare.i_create_a_local_model(self)
            prediction_create.i_create_a_prediction_op(
                self, example["input_data"], example["operating_point"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_prediction_op(
                self, example["input_data"], example["operating_point"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario3(self):
        """
        Scenario: Successfully comparing predictions for deepnets with operating point:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a deepnet with objective "<objective_id>" and "<model_conf>"
            And I wait until the deepnet is ready less than <model_wait> secs
            And I create a local deepnet
            When I create a prediction with operating point "<operating_point>" for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local prediction with operating point "<operating_point>" for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "model_conf",
                   "operating_point"]
        examples = [
            ['data/iris.csv', '10', '50', '60', '{"petal width": 4}', '000004',
             'Iris-setosa', '{}', {"kind": "probability", "threshold": 1,
                                   "positive_class": "Iris-virginica"}]]
        show_doc(self.test_scenario3)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_deepnet_with_objective_and_params(
                self, example["objective_id"], example["model_conf"])
            model_create.the_deepnet_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_compare.i_create_a_local_deepnet(self)
            prediction_create.i_create_a_deepnet_prediction_with_op(
                self, example["input_data"], example["operating_point"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_deepnet_prediction_with_op(
                self, example["input_data"], example["operating_point"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario4(self):
        """
        Scenario: Successfully comparing predictions in operating points for ensembles:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create an ensemble
            And I wait until the ensemble is ready less than <model_wait> secs
            And I create a local ensemble
            When I create a prediction for "<input_data>" in "<operating_point>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local ensemble prediction for "<input_data>" in "<operating_point>"
            Then the local ensemble prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "prediction", "operating_point",
                   "objective_id"]
        examples = [
            ['data/iris.csv', '10', '50', '50', '{"petal width": 4}',
             'Iris-setosa',
             {"kind": "probability", "threshold": 0.1,
              "positive_class": "Iris-setosa"}, "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal width": 4}',
             'Iris-virginica',
             {"kind": "probability", "threshold": 0.9,
              "positive_class": "Iris-setosa"}, "000004"],
            ['data/iris.csv', '10', '50', '50',
             '{"sepal length": 4.1, "sepal width": 2.4}',  'Iris-setosa',
             {"kind": "confidence", "threshold": 0.1,
              "positive_class": "Iris-setosa"}, "000004"],
            ['data/iris.csv', '10', '50', '50',
             '{"sepal length": 4.1, "sepal width": 2.4}', 'Iris-versicolor',
             {"kind": "confidence", "threshold": 0.9,
              "positive_class": "Iris-setosa"}, "000004"]]
        show_doc(self.test_scenario4)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            ensemble_create.i_create_an_ensemble(self, shared=example["data"])
            ensemble_create.the_ensemble_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            ensemble_create.create_local_ensemble(self)
            prediction_create.i_create_an_ensemble_prediction_op(
                self, example["input_data"], example["operating_point"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_ensemble_prediction_op(
                self, example["input_data"], example["operating_point"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario5(self):
        """
        Scenario: Successfully comparing predictions in operating kind for models:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a model
            And I wait until the model is ready less than <model_wait> secs
            And I create a local model
            When I create a prediction for "<input_data>" in "<operating_kind>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local prediction for "<input_data>" in "<operating_kind>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "prediction", "operating_kind",
                   "objective_id"]
        examples = [
            ['data/iris.csv', '10', '50', '50',
             '{"petal length": 2.46, "sepal length": 5}', 'Iris-versicolor',
             "probability", "000004"],
            ['data/iris.csv', '10', '50', '50',
             '{"petal length": 2.46, "sepal length": 5}', 'Iris-versicolor',
             "confidence", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2}',
             'Iris-setosa',  "probability", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2}',
             'Iris-setosa',  "confidence", "000004"]]
        show_doc(self.test_scenario5)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_model(self, shared=example["data"])
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            prediction_compare.i_create_a_local_model(self)
            prediction_create.i_create_a_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario6(self):
        """
        Scenario: Successfully comparing predictions for deepnets with operating kind:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a deepnet with objective "<objective_id>" and "<model_conf>"
            And I wait until the deepnet is ready less than <model_wait> secs
            And I create a local deepnet
            When I create a prediction with operating kind "<operating_kind>" for "<input_data>"
            Then the prediction for "<objective>" is "<prediction>"
            And I create a local prediction with operating point "<operating_kind>" for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "model_conf",
                    "operating_kind"]
        examples = [
            ['data/iris.csv', '10', '50', '60', '{"petal length": 2.46}',
             '000004', 'Iris-setosa', '{}', "probability"],
            ['data/iris.csv', '10', '50', '60', '{"petal length": 6}',
             '000004', 'Iris-versicolor', '{}', "probability"]]
        show_doc(self.test_scenario6)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_deepnet_with_objective_and_params(
                self, example["objective_id"], example["model_conf"])
            model_create.the_deepnet_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_compare.i_create_a_local_deepnet(self)
            prediction_create.i_create_a_deepnet_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_deepnet_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario7(self):
        """
        Scenario: Successfully comparing predictions in operating points for ensembles:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create an ensemble
            And I wait until the ensemble is ready less than <model_wait> secs
            And I create a local ensemble
            When I create a prediction for "<input_data>" in "<operating_kind>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local ensemble prediction for "<input_data>" in "<operating_kind>"
            Then the local ensemble prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "prediction", "operating_kind",
                   "objective_id"]
        examples = [
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2.46}',
             'Iris-versicolor',  "probability", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2}',
             'Iris-setosa',  "probability", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2.46}',
             'Iris-versicolor',  "confidence", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2}',
             'Iris-setosa',  "confidence", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 2.46}',
             'Iris-versicolor',  "votes", "000004"],
            ['data/iris.csv', '10', '50', '50', '{"petal length": 1}',
             'Iris-setosa',  "votes", "000004"]]
        show_doc(self.test_scenario7)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            ensemble_create.i_create_an_ensemble(self, shared=example["data"])
            ensemble_create.the_ensemble_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            ensemble_create.create_local_ensemble(self)
            prediction_create.i_create_an_ensemble_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_ensemble_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario8(self):
        """
        Scenario: Successfully comparing predictions for logistic regressions with operating kind:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <datase_wait> secs
            And I create a logistic regression with objective "<objective_id>"
            And I wait until the logistic regression is ready less than <model_wait> secs
            And I create a local logistic regression
            When I create a prediction with operating kind "<operating_kind>" for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local prediction with operating point "<operating_kind>" for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction",
                   "operating_kind"]
        examples = [
            ['data/iris.csv', '10', '50', '60', '{"petal length": 5}',
             '000004', 'Iris-versicolor', "probability"],
            ['data/iris.csv', '10', '50', '60', '{"petal length": 2}',
             '000004', 'Iris-setosa', "probability"]]
        show_doc(self.test_scenario8)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_logistic_model(
                self, shared=example["data"])
            model_create.the_logistic_model_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_compare.i_create_a_local_logistic_model(self)
            prediction_create.i_create_a_logistic_prediction_with_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_logistic_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario9(self):
        """
        Scenario: Successfully comparing predictions for logistic regressions with operating kind and supervised model:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a logistic regression with objective "<objective_id>"
            And I wait until the logistic regression is ready less than <model_wait> secs
            And I create a local supervised model
            When I create a prediction with operating kind "<operating_kind>" for "<input_data>"
            Then the prediction for "<objective>" is "<prediction>"
            And I create a local prediction with operating point "<operating_kind>" for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction",
                   "operating_kind"]
        examples = [
            ['data/iris.csv', '10', '50', '60', '{"petal length": 5}',
             '000004', 'Iris-versicolor', "probability"],
            ['data/iris.csv', '10', '50', '60', '{"petal length": 2}',
             '000004', 'Iris-setosa', "probability"]]
        show_doc(self.test_scenario9)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_logistic_model(
                self, shared=example["data"])
            model_create.the_logistic_model_is_finished_in_less_than(
                self, example["model_wait"], shared=example["data"])
            prediction_compare.i_create_a_local_supervised_model(
                self, model_type="logistic_regression")
            prediction_create.i_create_a_logistic_prediction_with_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_logistic_prediction_op_kind(
                self, example["input_data"], example["operating_kind"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])
            prediction_compare.i_create_a_local_bigml_model(
                self, model_type="logistic_regression")
            prediction_compare.i_create_a_local_bigml_model_prediction(
                self, example["input_data"], prediction_type="prediction",
                operating_kind=example["operating_kind"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario10(self):
        """
        Scenario: Successfully comparing predictions for linear regression:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a linear regression with objective "<objective_id>" and "<model_conf>"
            And I wait until the linear regression is ready less than <model_wait> secs
            And I create a local linear regression
            When I create a prediction for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local prediction for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "model_conf",
                    "operating_kind"]
        examples = [
            ['data/grades.csv', '10', '50', '60',
             '{"000000": 1, "000001": 1, "000002": 1}', '000005', 29.63024,
             '{"input_fields": ["000000", "000001", "000002"]}'],
            ['data/iris.csv', '10', '50', '60',
             '{"000000": 1, "000001": 1, "000004": "Iris-virginica"}',
             '000003', 1.21187,
             '{"input_fields": ["000000", "000001", "000004"]}'],
            ['data/movies.csv', '10', '50', '60', '{"000007": "Action"}',
             '000009', 4.33333, '{"input_fields": ["000007"]}'],
            ['data/movies.csv', '10', '50', '60', '{"000006": "1999"}',
             '000009', 3.28427, '{"input_fields": ["000006"], "bias": false}']]
        show_doc(self.test_scenario10)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            linear_create.i_create_a_linear_regression_with_objective_and_params(
                self, example["objective_id"], example["model_conf"])
            linear_create.the_linear_regression_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_compare.i_create_a_local_linear(self)
            prediction_create.i_create_a_linear_prediction(
                self, example["input_data"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_linear_prediction(
                self, example["input_data"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])

    def test_scenario11(self):
        """
        Scenario: Successfully comparing predictions for logistic regressions with operating point:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <source_wait> secs
            And I create a dataset
            And I wait until the dataset is ready less than <dataset_wait> secs
            And I create a logistic regression with objective "<objective_id>"
            And I wait until the logistic regression is ready less than <model_wait> secs
            And I create a local logistic regression
            When I create a prediction with operating point "<operating_point>" for "<input_data>"
            Then the prediction for "<objective_id>" is "<prediction>"
            And I create a local prediction with operating point "<operating_point>" for "<input_data>"
            Then the local prediction is "<prediction>"
        """
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction", "model_conf",
                   "operating_point"]
        examples = [
            ['data/iris.csv', '10', '50', '60', '{"petal width": 4}', '000004',
             'Iris-versicolor', '{"default_numeric_value": "mean"}',
             {"kind": "probability", "threshold": 1,
              "positive_class": "Iris-virginica"}]]
        show_doc(self.test_scenario11)
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_logistic_model_with_objective_and_parms(
                self, example["objective_id"], example["model_conf"])
            model_create.the_logistic_model_is_finished_in_less_than(
                self, example["model_wait"])
            prediction_compare.i_create_a_local_logistic_model(self)
            prediction_create.i_create_a_logistic_prediction_with_op(
                self, example["input_data"], example["operating_point"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])
            prediction_compare.i_create_a_local_prediction_op(
                self, example["input_data"], example["operating_point"])
            prediction_compare.the_local_prediction_is(
                self, example["prediction"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Testing local dataset transformations

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import compare_dataset_steps as dataset_compare


class TestLocalDataset:
    """Testing Local class for datasets"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario 1: Successfully creating a transformation from a local dataset in a json file:
            Given I create a local dataset from a "<dataset_file>" file
            Then the transformed data for "<input_data>" is "<output_data>"
        """
        show_doc(self.test_scenario1)
        headers = ["dataset_file", "input_data", "output_data"]
        examples = [
            ['bigml/tests/my_dataset/my_flatline_ds.json',
             '{"plasma glucose": 120, "age": 30, "bmi": 46}',
             '{"plasma glucose": 120, "age": 30, "glucose half": 60}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            dataset_compare.i_create_a_local_dataset_from_file(
                self, example["dataset_file"])
            dataset_compare.the_transformed_data_is(
                self, example["input_data"], example["output_data"])



# -*- coding: utf-8 -*-
#pylint: disable=abstract-method
#
# Copyright 2017-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base class for time series'' REST calls

   https://bigml.com/api/timeseries

"""

try:
    import simplejson as json
except ImportError:
    import json


from bigml.api_handlers.resourcehandler import ResourceHandlerMixin
from bigml.api_handlers.resourcehandler import check_resource_type, \
    resource_is_ready
from bigml.constants import TIME_SERIES_PATH


class TimeSeriesHandlerMixin(ResourceHandlerMixin):
    """This class is used by the BigML class as
       a mixin that provides the REST calls models. It should not
       be instantiated independently.

    """
    def __init__(self):
        """Initializes the TimeSeriesHandler. This class is intended
           to be used as a mixin on ResourceHandler, that inherits its
           attributes and basic method from BigMLConnection, and must not be
           instantiated independently.

        """
        self.time_series_url = self.url + TIME_SERIES_PATH

    def create_time_series(self, datasets,
                           args=None, wait_time=3, retries=10):
        """Creates a time series from a `dataset`
           of a list o `datasets`.

        """
        create_args = self._set_create_from_datasets_args(
            datasets, args=args, wait_time=wait_time, retries=retries)

        body = json.dumps(create_args)
        return self._create(self.time_series_url, body)

    def get_time_series(self, time_series, query_string='',
                        shared_username=None, shared_api_key=None):
        """Retrieves a time series.

           The model parameter should be a string containing the
           time series id or the dict returned by
           create_time_series.
           As a time series is an evolving object that is processed
           until it reaches the FINISHED or FAULTY state, the function will
           return a dict that encloses the time series
           values and state info available at the time it is called.

           If this is a shared time series, the username and
           sharing api key must also be provided.
        """
        check_resource_type(time_series, TIME_SERIES_PATH,
                            message="A time series id is needed.")
        return self.get_resource(time_series,
                                 query_string=query_string,
                                 shared_username=shared_username,
                                 shared_api_key=shared_api_key)

    def time_series_is_ready(self, time_series, **kwargs):
        """Checks whether a time series's status is FINISHED.

        """
        check_resource_type(time_series, TIME_SERIES_PATH,
                            message="A time series id is needed.")
        resource = self.get_time_series(time_series, **kwargs)
        return resource_is_ready(resource)

    def list_time_series(self, query_string=''):
        """Lists all your time series.

        """
        return self._list(self.time_series_url, query_string)

    def update_time_series(self, time_series, changes):
        """Updates a time series.

        """
        check_resource_type(time_series, TIME_SERIES_PATH,
                            message="A time series id is needed.")
        return self.update_resource(time_series, changes)

    def delete_time_series(self, time_series, query_string=''):
        """Deletes a time series.

        """
        check_resource_type(time_series, TIME_SERIES_PATH,
                            message="A time series id is needed.")
        return self.delete_resource(time_series, query_string=query_string)

    def clone_time_series(self, time_series,
                          args=None, wait_time=3, retries=10):
        """Creates a cloned time_series from an existing `time series`

        """
        create_args = self._set_clone_from_args(
            time_series, "timeseries", args=args, wait_time=wait_time,
            retries=retries)

        body = json.dumps(create_args)
        return self._create(self.time_series_url, body)



# -*- coding: utf-8 -*-
#
# Copyright 2021-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Declared exceptions.

"""

class ResourceException(Exception):
    """Base class to any exception that arises from a bad structured resource

    """


class NoRootDecisionTree(ResourceException):
    """The decision tree structure has no "root" attribute """



class FaultyResourceError(Exception):
    """Exception to be raised when retrieving a Faulty resource """



# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Miscellaneous utility functions for image processing

"""
import os
import tempfile

from PIL import Image

from bigml.constants import TEMP_DIR, TOP_IMAGE_SIZE as TOP_SIZE, DECIMALS


def resize_to(image, top_size=TOP_SIZE):
    """Resizing the image to a maximum width or height """
    width, height = image.size
    if width > top_size or height > top_size:
        if width > height:
            ratio = height / width
            image = image.resize((top_size , int(ratio * top_size)),
                                 Image.BICUBIC)
        else:
            ratio = width / height
            image = image.resize((int(ratio * top_size), top_size),
                                  Image.BICUBIC)
    return image


def to_relative_coordinates(image_file, regions_list):
    """Transforms predictions with regions having absolute pixels regions
    to the relative format used remotely and rounds to the same precision.
    """

    if regions_list:
        image_obj = Image.open(image_file)
        width, height = image_obj.size
        for index, region in enumerate(regions_list):
            [xmin, ymin, xmax, ymax] = region["box"]
            region["box"] = [round(xmin / width, DECIMALS),
                             round(ymin / height, DECIMALS),
                             round(xmax / width, DECIMALS),
                             round(ymax / height, DECIMALS)]
            region["score"] = round(region["score"], DECIMALS)
            regions_list[index] = region
    return regions_list


def remote_preprocess(image_file):
    """Emulating the preprocessing of images done in the backend to
    get closer results in local predictions
    """
    # converting to jpg
    image = Image.open(image_file)
    if not (image_file.lower().endswith(".jpg") or
            image_file.lower().endswith(".jpeg")):
        image = image.convert('RGB')
    # resizing to top size=512
    resize_to(image)
    with tempfile.NamedTemporaryFile(delete=False) as temp_fp:
        tmp_file_name = os.path.join(TEMP_DIR, "%s.jpg" % temp_fp.name)
        # compressing to 90%
        image.save(tmp_file_name, quality=90)
    return tmp_file_name



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2018-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating external connectors

"""
import json

from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_external_steps as connector_create

class TestExternalConnector:
    """Testing external connector creation"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating an external connector:
            Given I create an external connector from environment vars
            And I wait until the external connector is ready less than <conn_wait> secs
            And I update the external connector with args <args>
            And the external connector has arguments <args>
        """
        show_doc(self.test_scenario1)
        headers = ["conn_wait", "args"]
        examples = [
            ['20', '{"name": "my connector name"}']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            connector_create.i_create_external_connector(self)
            connector_create.the_external_connector_is_finished(
                self, example["conn_wait"])
            connector_create.i_update_external_connector_with(
                self, example["args"])
            connector_create.the_external_connector_is_finished(
                self, example["conn_wait"])
            connector_create.external_connector_has_args(
                example["args"])



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,line-too-long,attribute-defined-outside-init
#pylint: disable=locally-disabled,unused-import
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


""" Creating public model predictions

"""
from .world import world, setup_module, teardown_module, show_doc, \
    show_method
from . import create_source_steps as source_create
from . import create_dataset_steps as dataset_create
from . import create_model_steps as model_create
from . import create_prediction_steps as prediction_create
from . import compare_predictions_steps as compare_pred

class TestPublicModelPrediction:
    """Testing published models"""

    def setup_method(self, method):
        """
            Debug information
        """
        self.bigml = {}
        self.bigml["method"] = method.__name__
        print("\n-------------------\nTests in: %s\n" % __name__)

    def teardown_method(self):
        """
            Debug information
        """
        print("\nEnd of tests in: %s\n-------------------\n" % __name__)
        self.bigml = {}

    def test_scenario1(self):
        """
        Scenario: Successfully creating a prediction using a public model:
            Given I create a data source uploading a "<data>" file
            And I wait until the source is ready less than <time_1> secs
            And I create a dataset
            And I wait until the dataset is ready less than <time_2> secs
            And I create a model
            And I wait until the model is ready less than <time_3> secs
            And I make the model public
            And I wait until the model is ready less than <time_3> secs
            And I check the model status using the model's public url
            When I create a prediction for "<data_input>"
            Then the prediction for "<objective>" is "<prediction>"
        """
        show_doc(self.test_scenario1)
        headers = ["data", "source_wait", "dataset_wait", "model_wait",
                   "input_data", "objective_id", "prediction"]
        examples = [
            ['data/iris.csv', '10', '10', '10', '{"petal width": 0.5}',
             '000004', 'Iris-setosa']]
        for example in examples:
            example = dict(zip(headers, example))
            show_method(self, self.bigml["method"], example)
            source_create.i_upload_a_file(
                self, example["data"], shared=example["data"])
            source_create.the_source_is_finished(
                self, example["source_wait"], shared=example["data"])
            dataset_create.i_create_a_dataset(self, shared=example["data"])
            dataset_create.the_dataset_is_finished_in_less_than(
                self, example["dataset_wait"], shared=example["data"])
            model_create.i_create_a_model(self)
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            model_create.make_the_model_public(self)
            model_create.the_model_is_finished_in_less_than(
                self, example["model_wait"])
            model_create.model_from_public_url(self)
            prediction_create.i_create_a_prediction(
                self, example["input_data"])
            prediction_create.the_prediction_is(
                self, example["objective_id"], example["prediction"])



# -*- coding: utf-8 -*-
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Flatline: Class that encapsulates the Flatline expressions interpreter
"""

from javascript import require


class Flatline:
    """A bridge to an underlying nodejs Flatline interpreter.

    This class uses JSPyBridge to launch a Nodejs interpreter that loads
    Flatline's javascript implementation and allows interaction via
    Python constructs.

    Example:

      Flatline.check_lisp('(+ 1 2)')
      Flatline.check_json(["f", 0], dataset=dataset)

    """

    __FLATLINEJS = require('./flatline/flatline-node.js')
    interpreter = __FLATLINEJS.bigml.dixie.flatline

    #pylint: disable=locally-disabled,invalid-name
    @staticmethod
    def infer_fields(row, prefix=None, offset=None):
        """Utility function generating a mock list of fields.

        Usually, checks and applications of Flatline expressions run
        in the context of a given dataset's field descriptors, but
        during testing it's useful sometimes to provide a mock set of
        them, based on the types of the values of the test input rows.

        Example:

           In[1]: Interpreter.infer_fields([0, 'a label'])
           Out[2]: [{'column_number': 0,
                      'datatype': 'int64',
                      'id': '000000',
                      'optype': 'numeric'},
                     {'column_number': 1,
                      'datatype': 'string',
                      'id': '000001',
                      'optype': 'categorical'}]

        """
        result = []
        id_ = 0
        for v in row:
            t = type(v)
            optype = 'categorical'
            datatype = 'string'
            if (t is int or t is float):
                optype = 'numeric'
                if t is float:
                    datatype = 'float64'
                else:
                    datatype = 'int64'
            id_str = '%06x' % id_
            if prefix:
                length = len(prefix)
                id_str = prefix + id_str[length:]
            column = id_
            if offset:
                column = offset + id_
            result.append({'id': id_str,
                           'optype':optype,
                           'datatype': datatype,
                           'column_number': column})
            id_ = id_ + 1
        return result

    @staticmethod
    def _dataset(dataset, rows):
        """The dataset argument should be a Dataset that contains the
        in_fields information
        """
        try:
            return {"fields": dataset.in_fields}
        except AttributeError:
            if len(rows) > 0:
                return {'fields': Flatline.infer_fields(rows[0])}
            return None

    @staticmethod
    def defined_functions():
        """A list of the names of all defined Flaline functions"""
        return Flatline.interpreter.defined_primitives

    @staticmethod
    def check_lisp(sexp, fields=None):
        """Checks whether the given lisp s-expression is valid.

        Any operations referring to a dataset's fields will use the
        information found in fields structure.

        """
        r = Flatline.interpreter.evaluate_sexp(sexp, fields, True).valueOf()
        return r

    @staticmethod
    def check_json(json_sexp, fields=None):
        """Checks whether the given JSON s-expression is valid.

        Works like `check_lisp` (which see), but taking a JSON
        expression represented as a native Python list instead of a
        Lisp sexp string.

        """
        r = Flatline.interpreter.evaluate_js(json_sexp, fields).valueOf()
        return r

    @staticmethod
    def lisp_to_json(sexp):
        """ Auxliary function transforming Lisp to Python representation."""
        return Flatline.interpreter.sexp_to_js(sexp)

    @staticmethod
    def json_to_lisp(json_sexp):
        """ Auxliary function transforming Python to lisp representation."""
        return Flatline.interpreter.js_to_sexp(json_sexp)

    @staticmethod
    def apply_lisp(sexp, rows, dataset=None):
        """Applies the given Lisp sexp to a set of input rows.

        Input rows are represented as a list of lists of native Python
        values. The dataset info should be provided as a Dataset object.
        If no dataset is provided, the field characteristics
        of the input rows are guessed using `infer_fields`.

        """
        return Flatline.interpreter.eval_and_apply_sexp(
            sexp,
            Flatline._dataset(dataset, rows),
            rows)

    @staticmethod
    def apply_json(json_sexp, rows, dataset=None):
        """Applies the given JSON sexp to a set of input rows.

        As usual, JSON sexps are represented as Python lists,
        e.g. ["+", 1, 2].

        Input rows are represented as a list of lists of native Python
        values. The dataset info should be provided as a Dataset object.
        If no dataset is provided, the field characteristics
        of the input rows are guessed using `infer_fields`.

        """
        return Flatline.interpreter.eval_and_apply_js(
            json_sexp,
            Flatline._dataset(dataset, rows),
            rows)



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,cyclic-import
#
# Copyright 2022-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Pipeline: Classes that encapsulate the information needed to add the new
fields and predictions defined in a sequence of transformations or models.
The arguments to create a Pipeline are its name and the list of
datasets and models (and/or anomaly dectectors, clusters,
etc.) that describe the input data processing to be used.

"""

import os
import zipfile

from datetime import datetime

from bigml.api import get_api_connection, get_resource_id, get_resource_type
from bigml.util import use_cache, load, check_dir, get_data_format, \
    format_data, save_json, fs_cache_get, fs_cache_set, \
    dump, asciify
from bigml.constants import STORAGE
from bigml.dataset import Dataset
from bigml.supervised import SupervisedModel
from bigml.cluster import Cluster
from bigml.anomaly import Anomaly
from bigml.pca import PCA
from bigml.pipeline.transformer import BMLDataTransformer, DataTransformer

try:
    from bigml.topicmodel import TopicModel
    NO_TOPIC = False
except ImportError:
    NO_TOPIC = True


if NO_TOPIC:
    LOCAL_CLASSES = {
        "dataset": Dataset,
        "cluster": Cluster,
        "anomaly": Anomaly,
        "pca": PCA,
        }
else:
    LOCAL_CLASSES = {
        "dataset": Dataset,
        "cluster": Cluster,
        "anomaly": Anomaly,
        "topicmodel": TopicModel,
        "pca": PCA,
        }


def get_datasets_chain(dataset, dataset_list=None):
    """Builds recursively the chain of datasets leading to a dataset """
    if dataset_list is None:
        dataset_list = []
    dataset_list.append(dataset)
    if dataset.origin_dataset is None:
        return dataset_list

    return get_datasets_chain(dataset.origin_dataset, dataset_list)


def get_datasets_dict(dataset, dataset_dict=None):
    """Stores a dictionary dataset_id -> Dataset for the chain of datasets """
    if dataset_dict is None:
        dataset_dict = {}
    dataset_dict.update({dataset.resource_id: dataset})
    if dataset.origin_dataset is None:
        return dataset_dict

    return get_datasets_dict(dataset.origin_dataset, dataset_dict)


def check_in_path(path, resource_list):
    """Checks whether a list of resources is stored in a folder """
    for resource_id in resource_list:
        if not os.path.exists(os.path.join(
                path, resource_id.replace("/", "_"))):
            return False
    return True


class Pipeline(DataTransformer):
    """Class to define sequential transformations. The transformations can
    come from BigML resources or be defined as Pipe steps defined as functions
    to be applied to DataFrame pipes, scikit pipelines

    """
    def __init__(self, name, steps=None, resource_id=None, description=None):
        """Builds a Pipeline from the list of steps provided in the `steps`
        argument. It is compulsory to assign a name that will be used as
        reference
        :param name: Reference name for the pipeline
        :type name: str
        :param steps: List of DataTransformers. All of them need to offer a
                      `.transform` method
        :type steps: list
        :param description: Description of the transformations in the pipeline
        :type description: str
        """
        super().__init__(None, # no generator is provided
             None, # no data format is assumed
             resource_id or name,
             name,
             description)

        self.steps = []
        self.extend(steps)

    def extend(self, steps=None):
        """Adding new transformations to the Pipeline steps"""
        if steps is None:
            steps = []
        for step in steps:
            if not hasattr(step, "transform"):
                raise ValueError("Failed to find the .transform method in "
                    "all the Pipeline steps.")
        self.steps.extend(steps)

    def transform(self, input_data_list, out_format=None):
        """Applying the Pipeline transformations and predictions on the
        list of input data. `out_format` forces the output format
        to either a DataFrame or a list of dictionaries.

        """
        result = self.data_transform(input_data_list)
        if out_format is not None:
            current_format = get_data_format(result)
            if current_format != out_format:
                return format_data(result, out_format)
        return result

    def data_transform(self, input_data_list):
        """Delegates transformation to each DataTransformer step"""
        current_format = get_data_format(input_data_list)
        if len(self.steps) == 0:
            return input_data_list
        inner_data_list = input_data_list
        for index, step in enumerate(self.steps[:-1]):
            try:
                inner_data_list = step.transform(inner_data_list)
            except Exception as exc:
                raise ValueError(
                    "Failed to apply step number %s in pipeline %s: %s" %
                    (index, self.name, exc))
        try:
            inner_data_list = self.steps[-1].transform(
                inner_data_list, out_format=current_format)
            if hasattr(self.steps[-1], "add_input") and \
                    self.steps[-1].add_input:
                self.steps[-1].merge_input_data(
                    input_data_list, inner_data_list,
                    out_format=current_format)
        except Exception as exc:
            raise ValueError("Failed to apply the last step: %s" % exc)
        return inner_data_list


class BMLPipeline(Pipeline):
    """The class represents the sequential transformations (and predictions)
    that the input data goes through in a prediction workflow.
    Reproduces the pre-modeling steps that need to be applied before
    the application of the model predict (centroid, anomaly score, etc.)
    method to add the final prediction. The mandatory arguments for the class
    are:
      - name: Each pipeline needs to be identified with a unique name
      - resource_list: A list of resource IDs. Only datasets and supervised
                       or unsupervised model resources are allowed.

    When a dataset is provided, only the chain of transformations leading to
    that dataset structure is applied. When a model is provided, the input
    data is pre-modeled using that chain of transformations and the result
    is used as input for the predict-like method of the model, that adds the
    prediction to the result. If the pipeline is expected to use strictly
    the resources in the original resource_list, you can use the last_step
    argument

    """
    def __init__(self, name, resource_list=None, description=None, api=None,
                 cache_get=None, init_settings=None, execution_settings=None,
                 last_step=False):
        """The pipeline needs
        :param name: A unique name that will be used when caching the
                     resources it needs to be executed.
        :type name: str
        :param resource_list: A dataset/model ID or a list of them
                              to define the transformations and predictions
                              to be added to the input data.
        :type resource_list: list
        Optionally, it can receive:
        :param description: A description of the pipeline procedure
        :type description: str
        :param api: A BigML API connection object
        :type api: BigML
        :param cache_get: A cache_get function to retrieve cached resources
        :type cache_get: function
        :param init_settings: A dictionary describing the optional arguments
                              added when instantiating the local model
                              (one per model ID)
           e.g.:
              {"deepnet/111111111111111111": {
                  "operation_settings": {
                      "region_score_threshold": 0.6}},
               "deepnet/222222222222222222": {
                  "operation_settings": {
                      "region_score_threshold": 0.7}}}
        :type init_settings: dict
        :param execution_settings: A dictionary describing the optional
                                   arguments added when creating the
                                   predictions.
           e.g.:
              {"model/111111111111111111": {
                  "missing_strategy": 1},
               "model/222222222222222222": {
                  "operating_kind": "confidence"}}
        :type execution_settings: dict

        """

        if resource_list is None and use_cache(cache_get):
            self.__dict__ = load(name, cache_get)
        else:
            super().__init__(name, description=description)

            # API related attributes
            if resource_list is None:
                resource_list = []
            self.resource_list = resource_list
            if isinstance(resource_list, str):
                self.resource_list = [resource_list]
            for item in self.resource_list:
                resource_id = get_resource_id(item)
                if resource_id is None:
                    raise ValueError("Only resource IDs are allowed as first "
                        "argument.")
            self.init_settings = init_settings or {}
            self.execution_settings = execution_settings or {}
        self._api = get_api_connection(api)
        if self._api.storage is None:
            self._api.storage = self._get_pipeline_storage()
        self._cache_get = cache_get
        self.steps = []
        self.extend(self.__retrieve_steps(last_step))

    def __retrieve_steps(self, last_step):
        """Retrieving the steps that need to be used to reproduce the
        transformations leading to the resources given in the original list
        """
        local_resources = []
        init_settings = self.init_settings.copy()
        execution_settings = self.execution_settings.copy()
        datasets = {}
        steps = []

        kwargs = {}
        if self._api is not None:
            kwargs["api"] = self._api
        if self._cache_get is not None:
            kwargs["cache_get"] = self._cache_get

        for resource_id in self.resource_list:
            init_settings[resource_id] = init_settings.get(
                resource_id, {})
            init_settings[resource_id].update(kwargs)

        for index, resource in enumerate(self.resource_list):
            resource_id = get_resource_id(resource)
            resource_type = get_resource_type(resource_id)
            local_class = LOCAL_CLASSES.get(resource_type, SupervisedModel)
            kwargs = init_settings.get(resource_id, {})
            local_resource = local_class(resource, **kwargs)
            if isinstance(local_resource, SupervisedModel):
                execution_settings[resource_id] = \
                    execution_settings.get(
                        resource_id, {})
                execution_settings[resource_id].update({"full": True})
            local_resources.append([local_resource])
            if (hasattr(local_resource, "parent_id") and \
                    get_resource_type(local_resource.parent_id) == "dataset"):
                if local_resource.parent_id in datasets:
                    dataset = datasets[local_resource.parent_id]
                else:
                    dataset = Dataset(local_resource.parent_id,
                                      api=self._api)
                    datasets = get_datasets_dict(dataset, datasets)
                if not last_step:
                    dataset_chain = get_datasets_chain(dataset)
                    local_resources[index].extend(dataset_chain)
                    local_resources[index].reverse()

        try:
            new_resources = local_resources[0][:]
        except IndexError:
            new_resources = []
        for index, resources in enumerate(local_resources):
            if index < 1:
                continue
            for resource in resources:
                if resource not in new_resources:
                    new_resources.append(resource)
        local_resources = new_resources
        for local_resource in local_resources:
            # non-flatline datasets will not add transformations
            if isinstance(local_resource, Dataset) and \
                    local_resource.origin_dataset is not None and \
                    local_resource.transformations is None:
                continue
            execution_settings = self.execution_settings.get(
                local_resource.resource_id, {})
            steps.append(BMLDataTransformer(
                local_resource, **execution_settings))
        return steps

    def _get_pipeline_storage(self):
        """ Creating a separate folder inside the given storage folder to
        contain the pipeline related models based on the pipeline name.
        If the folder already exists, first we check that all the resources
        in the resources list are already stored there. If that's not the
        case, we rename the folder by adding a datetime suffix and create a
        new pipeline folder to store them.
        """
        if self._api.storage is None:
            self._api.storage = STORAGE
        path = os.path.join(self._api.storage, self.name)
        if os.path.exists(path):
            if check_in_path(path, self.resource_list):
                return path
            # adding a suffix to store old pipeline version
            datetime_str = str(datetime.now()).replace(" ", "_")
            bck_path = f"{path}_{datetime_str}_bck"
            os.rename(path, bck_path)
        check_dir(path)
        return path

    def export(self, output_directory=None):
        """Exports all the resources needed in the pipeline to the user-given
        output directory. The entire pipeline folder is exported and its name
        is used as filename.
        """
        def zipdir(path, ziph):
            # ziph is zipfile handle
            for root, _, files in os.walk(path):
                for file in files:
                    ziph.write(os.path.join(root, file),
                               os.path.relpath(os.path.join(root, file),
                                               os.path.join(path, '..')))

        if output_directory is None:
            output_directory = os.getcwd()
        check_dir(output_directory)
        name = asciify(self.name)
        out_filename = os.path.join(output_directory, f"{name}.zip")

        # write README file with the information that describes the Pipeline
        name = self.name
        description = self.description or ""
        resources = ", ".join(self.resource_list)
        readme = (f"Pipeline name: {name}\n{description}\n\n"
                  f"Built from: {resources}")
        with open(os.path.join(self._api.storage, "README.txt"), "w",
                  encoding="utf-8") as readme_handler:
            readme_handler.write(readme)
        # write JSON file describing the pipeline resources
        pipeline_vars = vars(self)
        stored_vars = {}
        for key, value in pipeline_vars.items():
            if not key.startswith("_") and not key == "steps":
                stored_vars.update({key: value})
        pipeline_filename = os.path.join(self._api.storage, asciify(self.name))
        save_json(stored_vars, pipeline_filename)
        with zipfile.ZipFile(out_filename, 'w', zipfile.ZIP_DEFLATED) as zipf:
            zipdir(self._api.storage, zipf)

    def dump(self, output_dir=None, cache_set=None):
        """Uses msgpack to serialize the resource object and all its steps
        If cache_set is filled with a cache set method, the method is called
        to store the serialized value
        """
        pipeline_vars = vars(self)
        stored_vars = {}
        for key, value in pipeline_vars.items():
            if not key.startswith("_") and not key == "steps":
                stored_vars.update({key: value})
        if output_dir is not None:
            check_dir(output_dir)
            cache_set = cache_set or fs_cache_set(output_dir)
        dump(stored_vars, output=None, cache_set=cache_set)
        for step in self.steps:
            step.dump(cache_set=cache_set)

    @classmethod
    def load(cls, name, dump_dir):
        """Restores the information of the pipeline and its steps from a
        previously dumped pipeline file. The objects used in each step
        of the pipeline are expected to be in the same
        """
        if dump_dir is not None and name is not None:
            return cls(name,
                       None,
                       cache_get=fs_cache_get(dump_dir))
        return None



# -*- coding: utf-8 -*-
#pylint: disable=locally-disabled,unused-argument,no-member
#
# Copyright 2015-2023 BigML
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from bigml.api import HTTP_CREATED, HTTP_ACCEPTED
from bigml.api import FINISHED, FAULTY
from bigml.execution import Execution

from .read_resource_steps import wait_until_status_code_is
from .world import world, eq_


def the_execution_and_attributes(step, param, param_value, result):
    """Step: the script id is correct, the value of <param> is <param_value>
    and the result is <result>
    """
    eq_(world.script['resource'], world.execution['script'])
    eq_(world.execution['execution']['results'][0], result)
    res_param_value = world.execution[param]
    eq_(res_param_value, param_value,
        ("The execution %s is %s and the expected %s is %s" %
         (param, param_value, param, param_value)))


def the_execution_ids_and_attributes(step, number_of_scripts,
                                     param, param_value, result):
    """Step: the script ids are correct, the value of <param> is <param_value>
    and the result is <result>
    """
    scripts = world.scripts[-number_of_scripts:]
    eq_(scripts, world.execution['scripts'])
    eq_(world.execution['execution']['results'], result)
    res_param_value = world.execution[param]
    eq_(res_param_value, param_value,
        ("The execution %s is %s and the expected %s is %s" %
         (param, param_value, param, param_value)))


def i_create_an_execution(step):
    """Step: I create a whizzml execution from an existing script"""
    resource = world.api.create_execution(world.script['resource'],
                                          {"project": world.project_id})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.execution = resource['object']
    world.executions.append(resource['resource'])


def i_create_an_execution_from_list(step, number_of_scripts=2):
    """Step: I create a whizzml execution from the last two scripts"""
    scripts = world.scripts[-number_of_scripts:]
    resource = world.api.create_execution(scripts,
                                          {"project": world.project_id})
    world.status = resource['code']
    eq_(world.status, HTTP_CREATED)
    world.location = resource['location']
    world.execution = resource['object']
    world.executions.append(resource['resource'])


def i_update_an_execution(step, param, param_value):
    """Step: I update the execution with <param>, <param_value>"""
    resource = world.api.update_execution(world.execution['resource'],
                                          {param: param_value})
    world.status = resource['code']
    eq_(world.status, HTTP_ACCEPTED)
    world.location = resource['location']
    world.execution = resource['object']


def wait_until_execution_status_code_is(step, code1, code2, secs):
    """Step: I wait until the execution status code is either <code1> or
    <code2> less than <secs>"""
    world.execution = wait_until_status_code_is(
        code1, code2, secs, world.execution)


def the_execution_is_finished(step, secs):
    """Steps: I wait until the script is ready less than <secs>"""
    wait_until_execution_status_code_is(step, FINISHED, FAULTY, secs)


def create_local_execution(step):
    """Step: I create a local execution"""
    step.bigml["local_execution"] = Execution(world.execution)


def the_local_execution_result_is(step, result):
    """Step: And the local execution result is <result>"""
    eq_(step.bigml["local_execution"].result, result)
